/* pgm.c  to do with robot programs
 *  18 Jan 96
 *  7 Mar 96
 *  update 23 Feb 2006
 *  P. Chongstitvatana
 */

#include "gp.h"

acell cell[CELLSIZE*MAXCELL];   /* storage pool for program */
int   freecell,top,bottom,emark;

/*-------------------  cell storage organisation

             ------------   ^
             | reserved |   |
             ------------   |
 freecell--> |          |  maxcell*cellsize
             |          |   |
   ncell --> |          |   |
             ------------   v


straight case reclaim

             -----------
             |  res    |
             -----------
 ----------> |  used   | <-- top
 |   end --> |         |
 |           -----------
 | emark --> |         | <-- freecell
 |           |         |
 ----------- |         | <-- bottom
  new link   -----------


split case reclaim

             -----------
		     |   res   |
		     -----------
             |  used   |
             |         |
             -----------
             |         | <-- emark
             -----------
             |  used   | <-- top
             -----------

---------------------- */

void initcell(void)   /* initialise links and freecell */
{
  int i;

  top = RESERVED*CELLSIZE;
  bottom = CELLSIZE*(MAXCELL-1);
  for( i=top; i<bottom; i+=CELLSIZE)
    cell[i] = i + CELLSIZE;
  cell[bottom] = NIL;
  freecell = top;
}

PRIVATE void error(char *msg)
{
  fprintf(stderr,"%s\n",msg);
  exit(-1);
}
/* return a new cell pointer */
PRIVATE int getacell(void)
{
  int ap,i;

  ap = freecell;
  if (ap == NIL) error( "no more cell");
  freecell = cmd(ap);
  /* NIL out the new cell */
  for( i=0; i<CELLSIZE; i++) arg(i,ap) = NIL;
  return ap;
}

void markcell(void)
{
  emark = freecell;
}

/* reclaim the cell from top to end
 * there are two cases :
 * 1) straight, cells are contiguous
 * 2) spilt, cells are wrapper around, top <end, reclaim two seperate blocks
 *    and link them together
 */
void reclaim(void)
{
  int offset,ncell,end,i;

  offset = CELLSIZE*RESERVED;
  ncell = CELLSIZE*(MAXCELL-1);
  end = ( emark - CELLSIZE < offset ) ? ncell : emark - CELLSIZE ;
  if( top < end ) {    /* straight collect */
    for( i=top; i<end; i+=CELLSIZE )
      cell[i] = i + CELLSIZE;
  }else {                 /* split collect */
    for( i=top; i<ncell; i+=CELLSIZE )
      cell[i] = i + CELLSIZE;
    for( i=offset; i<end; i+=CELLSIZE )
      cell[i] = i + CELLSIZE;
    cell[ncell] = offset;
  }
  cell[end] = NIL;
  cell[bottom] = top;
  top = emark;
  bottom = end;
}

/* ------------------- generate population -----------------------*/

PRIVATE int genprog1(void) /* random prog size 1 cell */
{
  int ap;

  ap = getacell();
  cmd(ap) = myrnd(3) + IFAND;
  arg1(ap) = myrnd(IFAND);
  arg2(ap) = myrnd(IFAND);
  arg3(ap) = myrnd(IFAND);
  if( is4arg(cmd(ap)) )  arg4(ap) = myrnd(IFAND);
  return ap;
}
/* generate a random program */
/* depth is used to track the level, start with n */
PRIVATE int genprog(int depth)
{
  int ap, c, i;

  if ( depth == 0 ) return myrnd(IFAND); /* terminals only */
  c = myrnd(ENDCODE);
  if ( isterm(c) ) return c;
  ap = getacell();
  cmd(ap) = c;
  arg1(ap) = genprog(depth-1);
  arg2(ap) = genprog(depth-1);
  arg3(ap) = genprog(depth-1);
  if( is4arg(cmd(ap)) ) arg4(ap) = genprog(depth-1);
  return ap;
}
/* find length of a prog, (the number of symbol) */
int proglen(int ap)
{
  int len;

  if( isterm(ap) ) return 1;
  len = 1 + proglen(arg1(ap)) + proglen(arg2(ap)) + proglen(arg3(ap));
  if ( is4arg(cmd(ap)) ) len += proglen(arg4(ap));
  return len;
}
/* gen random prog length > n, return length in len */
int genprogn(int n, int *len)
{
  int ap, k;

  /* depth limit : n < 10 = 2, n < 20 = 3 etc. */
  k = n / 10 + 2;
  while ( 1 ) {
    ap = genprog(k);
    *len = proglen(ap);
    if (*len > n) break;
  }
  return ap;
}
/* output a compact form of a program */
void tofile(int ap, FILE *fp)
{
  if ( ap == NIL ) return;
  if (isterm(ap)) {
    fputc(ap+'a',fp);
    return;
  }
  fputc(cmd(ap)+'a',fp);
  tofile(arg1(ap),fp);
  tofile(arg2(ap),fp);
  tofile(arg3(ap),fp);
  if( is4arg(cmd(ap)) ) tofile(arg4(ap),fp);
}
/* input a compact form of a program */
int fromfile(FILE *fp)
{
  int c,ap;

  c = fgetc(fp);
  if ( c == EOF || c < 'a' || c > 'n' ) return NIL;
  c -= 'a';
  if ( isterm(c) ) return c;
  ap = getacell();
  cmd(ap) = c;
  arg1(ap) = fromfile(fp);
  arg2(ap) = fromfile(fp);
  arg3(ap) = fromfile(fp);
  if( is4arg(cmd(ap)) ) arg4(ap) = fromfile(fp);
  return ap;
}

/* ------------------- genetic operations ---------- */

#define  tosscoin()  (myrnd(3) == 0)  /* is true with prob 1/3 */

/* copy program tree, return pointer to new tree */
int copyprog(int ap)
{
  int ap2;

  if ( isterm(ap) ) return ap;
  ap2 = getacell();
  cmd(ap2) = cmd(ap);
  arg1(ap2) = copyprog(arg1(ap));
  arg2(ap2) = copyprog(arg2(ap));
  arg3(ap2) = copyprog(arg3(ap));
  if( is4arg(cmd(ap)) ) arg4(ap2) = copyprog(arg4(ap));
  return ap2;
}

/* find pointer to a  n th compenent of ap, return yes/no and ptr
 * ptr is undef if not find
 */
PRIVATE int findnode(int ap, int *n, int *ptr)
{
  int i,k,c;

  if ( *n == 0 ) { *ptr = ap; return 1; }
  c = cell[ap];
  if ( isterm(c) ) { (*n)--; return 0;}
  if ( c < ENDCODE ) {
    (*n)--;
    k = argsize(c);
    for( i=1; i<=k; i++ )
      if ( findnode( ap+i, n, ptr) ) return 1;
    return 0;
  }
  return  findnode( c, n, ptr);
}
/* swap two pgms at two random positions */
void crossover(int ap1, int ap2)
{
  int p1,p2,a,b,temp;

  p1 = myrnd(proglen(ap1)-1) + 1;
  p2 = myrnd(proglen(ap2)-1) + 1;
  a = ap1; b = ap2;
  findnode(ap1,&p1, &a);
  findnode(ap2,&p2, &b);
  temp = cell[a];
  cell[a] = cell[b];
  cell[b] = temp;
}
/* add a random prog at root */
int addprog(int ap)
{
  int ap2,i,a;

  ap2 = genprog1();
  i = myrnd(proglen(ap2) -1) +1;
  a = ap;
  if( findnode(ap2, &i, &a) ) cell[a] = ap;
  return ap2;
}
/* extend a terminal node with a random prog */
int extendprog(int ap)
{
  int ap2,k,i,a,s;

  ap2 = genprog1();
  k = proglen(ap) -1;
  do {
    i = myrnd(k)+1;
    a = ap;
    findnode(ap, &i, &a);
  } while ( ! (s = isterm(cell[a])) );
  if( s ) cell[a] = ap2;
  return ap;
}
/* create a mutant with prob according to tosscoin() */
int mutateprog(int ap)
{
  int ap2, b, c;

  if( isterm(ap) ) return (tosscoin() ? myrnd(IFAND) : ap);
  ap2 = getacell();
  c = tosscoin() ? myrnd(3) + IFAND : cmd(ap);
  cmd(ap2) = c;
  arg1(ap2) = mutateprog(arg1(ap));
  arg2(ap2) = mutateprog(arg2(ap));
  arg3(ap2) = mutateprog(arg3(ap));
  if( is4arg(c) )
    arg4(ap2) = is4arg(cmd(ap)) ? mutateprog(arg4(ap)) : genprogn(2,&b);
  return ap2;
}


/* ----------   useful tools --------------------- */

/* report no. of freecell, use to test reclaim() */
void countcell(void)
{
  int ap,cnt;

  cnt = 0;  ap = freecell;
  while( cmd(ap) != NIL ) {
    cnt++;
    ap = cmd(ap);
  }
  printf("%d cell\n",cnt);
}

/* printname of commands */

char printname[][8] = {
 "s+", "s-", "e+", "e-", "w+", "w-", "hit?", "see?", "inc?", "dec?", "out?",
 "if-and", "if-or", "if-not","$", 0 };

void printprog(int ap)
{
  int i,k;

  if (ap == NIL) return;
  if ( isterm(ap) ) {
    printf("%s ", printname[ap] );
    return;
  }
  printf("(%s ", printname[cmd(ap)] );
  k = argsize(cmd(ap));
  for( i=1; i<=k; i++) printprog(arg(i,ap));
  printf(")");
}

void printfromfile(char *fname)
{
  FILE *fp;
  int ap;

  initcell();
  fp = fopen(fname,"r");
  ap = fromfile(fp);
  fclose(fp);
  printf("pgm %s len %d\n",fname,proglen(ap));
  printprog(ap); nl();
}
/* test if two programs are similar , tree-equal */
int eqprog(int ap1, int ap2)
{
  int i,k;

  if ( isterm(ap1) )  return ( ap1 == ap2 );
  k = argsize(cmd(ap1));
  for( i=1; i<=k; i++)
    if (! eqprog(arg(i,ap1), arg(i,ap2) ) ) return 0;
  return 1;
}



